//
// $Id: ibfetch.cpp 1014 2011-04-10 04:41:36Z kos $
//

#include <limits.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <libgen.h>
#include <sys/time.h>
#include <stdarg.h>

#include <iostream>
#include <sstream>
#include <string>
#include <algorithm>
#include <cctype>
#include <exception>
#include <stdexcept>
#include <memory>
#include <map>
#include <sstream>

#include <log4cxx/logger.h>
#include <log4cxx/fileappender.h>
#include <log4cxx/asyncappender.h>
#include <log4cxx/rollingfileappender.h>
#include <log4cxx/simplelayout.h>
#include <log4cxx/patternlayout.h>

#include <boost/shared_ptr.hpp>

#include "Contract.h"
#include "Order.h"
#include "EPosixClientSocket.h"
#include "EPosixClientSocketPlatform.h"
#include "EWrapper.h"
#include "EClient.h"

#include "constants.h"
#include "util.h"
#include "csvwriter.h"
#include "logger.h"
#include "ibfetch.h"

using namespace std;

// own
IBFetch::IBFetch(LoggerRP& logger) : _pclient(new EPosixClientSocket(this)),
				     have_details(false),
				     _num_requests(0),
				     _completed_requests(0),
				     _failed_requests(0),
				     realtimeTicker(1),
				     realtimeCounter(0), 
				     _logger(logger) {
  _clientid = (int)getpid() + rand();
}

IBFetch::~IBFetch() { }

// trivial request management
void
IBFetch::request_new() {
  _num_requests++;
}

void
IBFetch::set_clientid(int id) {
  _clientid = id;
}

void
IBFetch::request_fail() {
  _failed_requests++;
  _completed_requests++;
}

void
IBFetch::request_done() {
  _completed_requests++;
}

// EWrapper implementation
void IBFetch::tickOptionComputation(TickerId tickerId, TickType tickType, double impliedVol, double delta,
				    double modelPrice, double pvDividend) {}
void IBFetch::tickGeneric(TickerId tickerId, TickType tickType, double value) {}
void IBFetch::tickString(TickerId tickerId, TickType tickType, const IBString& value) {}
void IBFetch::tickEFP(TickerId tickerId, TickType tickType, double basisPoints, const IBString& formattedBasisPoints,
		     double totalDividends, int holdDays, const IBString& futureExpiry, double dividendImpact, double dividendsToExpiry) {}
void IBFetch::orderStatus(OrderId orderId, const IBString &status, int filled,
			  int remaining, double avgFillPrice, int permId, int parentId,
			  double lastFillPrice, int clientId, const IBString& whyHeld) {}
void IBFetch::openOrderEnd() {}
void IBFetch::winError(const IBString &str, int lastError) {}
void IBFetch::updateAccountValue(const IBString& key, const IBString& val,
				const IBString& currency, const IBString& accountName) {}
void IBFetch::updateAccountTime(const IBString& timeStamp) {}
void IBFetch::nextValidId(OrderId orderId) {}
void IBFetch::accountDownloadEnd(const IBString& accountName) {}

void IBFetch::bondContractDetails(int reqId, const ContractDetails& contractDetails) {}
void IBFetch::contractDetailsEnd(int reqId) {}
void IBFetch::execDetails(int reqId, const Contract& contract, const Execution& execution) {}
void IBFetch::execDetailsEnd(int reqId) {}
void IBFetch::updateMktDepth(TickerId id, int position, int operation, int side,
			    double price, int size) {}
void IBFetch::updateMktDepthL2(TickerId id, int position, IBString marketMaker, int operation,
			      int side, double price, int size) {}
void IBFetch::updateNewsBulletin(int msgId, int msgType, const IBString& newsMessage, const IBString& originExch) {}
void IBFetch::managedAccounts(const IBString& accountsList) {}
void IBFetch::receiveFA(faDataType pFaDataType, const IBString& cxml) {}
void IBFetch::scannerParameters(const IBString &xml) {}
void IBFetch::scannerData(int reqId, int rank, const ContractDetails &contractDetails,
			 const IBString &distance, const IBString &benchmark, const IBString &projection,
			 const IBString &legsStr) {}
void IBFetch::scannerDataEnd(int reqId) {}
void IBFetch::realtimeBar(TickerId reqId, long time, double open, double high, double low, double close,
			 long volume, double wap, int count) {}
void IBFetch::fundamentalData(TickerId reqId, const IBString& data) {}
void IBFetch::deltaNeutralValidation(int reqId, const UnderComp& underComp) {}
void IBFetch::tickSnapshotEnd(int reqId) {}

void IBFetch::connectionClosed() { 
  int out = requestsOutstanding();
  if(out)
    _logger->log(Logger::WARN, "client disconnecting while outstanding requests = %d", out);
  _failed_requests += out;
  _completed_requests = _num_requests;
}

void 
IBFetch::updatePortfolio(const Contract& contract, int position,
			double marketPrice, double marketValue, double averageCost,
			double unrealizedPNL, double realizedPNL, const IBString& accountName) {
}

void 
IBFetch::openOrder(OrderId orderId, const Contract& c, const Order& o, const OrderState& ostate) {    
}

void 
IBFetch::error(const int id, const int errorCode, const IBString errorString) {
  if(id < 0)
    _logger->log(Logger::INFO, "[%d,%d]: %s", id, errorCode, errorString.c_str());
  else
    _logger->log(Logger::ERROR, "[%d,%d]: %s", id, errorCode, errorString.c_str());
  
  // code 165 tells us the "farm" is coming online. Anything else is a real error.
  if(id>0 && errorCode!=165)
    request_fail();
}

void
IBFetch::currentTime(long time) { 
  _logger->log(Logger::INFO, "Current time: %ld", time);
}

int
IBFetch::processReplies(bool drain, int timeout=1) throw(runtime_error) {
  fd_set fds;
  int fd = _pclient->fd();
  int nmesg = 0;
  struct timeval tv;
  int retval;

  do {
    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    FD_ZERO(&fds);
    FD_SET(fd, &fds);

    if(timeout)
      retval = select(fd+1, &fds, NULL, NULL, &tv);
    else
      retval = select(fd+1, &fds, NULL, NULL, NULL);

    if(retval<0)
      break;

    if(FD_ISSET(fd, &fds)) {
      if(!_pclient->checkMessages())
	throw(runtime_error("Failed to retrieve response: socket read error"));
    } else {
      break;
    }
    nmesg++;
  } while(drain);
  return nmesg;
}

void
IBFetch::contractDetails(int reqId, const ContractDetails& contractDetails) {
  details = contractDetails; 
  have_details = true;
  request_done();
}

// FX format: XYZ.ABC
// Length is always 7 including separator
//
// Futures format: RRMY[.EXCHANGE]
// R = root (e.g. ES, NQ); may be >=2 chars long
// M = month: March:H, June:M, Sept: U, Dec: Z
// Y = year: 2010:0, 2011:1, 2012: 2 etc

bool
IBFetch::resolveContract(string& spec, string& exchange, string& currency, string& date, Contract& contract) {
  const char numbers[] = "0123456789";
  size_t dotsep, numsep;
  static int request_id = 1121;

  dotsep = spec.find(".");
  numsep = spec.find_last_of(numbers);
  if(dotsep != string::npos && spec.length() == 7) {
    // FX
    transform(spec.begin(), spec.end(), spec.begin(), ::toupper);
    contract.symbol = spec.substr(0, dotsep);
    contract.localSymbol = spec;
    contract.secType = "CASH";
    if(exchange.size())
      contract.exchange = exchange;
    else
      contract.exchange = "IDEALPRO";
    contract.currency = spec.substr(dotsep+1);
  } else if(numsep != string::npos && spec.length() >= 4) {
    // futures
    contract.secType = "FUT";
    if(currency.size())
      contract.currency = currency;
    else
      contract.currency = "USD";
    contract.includeExpired = true;
    if(dotsep == string::npos) {
      transform(spec.begin(), spec.end()-2, spec.begin(), ::toupper);
      string newspec;
      resolve_futures_contract(spec, date, newspec);
      _logger->log(Logger::INFO, "resolved futures spec %s to: %s", spec.c_str(), newspec.c_str());
      contract.symbol = newspec.substr(0, newspec.length()-2);
      contract.localSymbol = newspec;
      if(exchange.size())
	contract.exchange = exchange;
      else
	contract.exchange = "GLOBEX";
    } else {
      transform(spec.begin(), spec.end(), spec.begin(), ::toupper);
      contract.symbol = spec.substr(0, dotsep-2);
      contract.localSymbol = spec.substr(0, dotsep);
      contract.exchange = spec.substr(dotsep+1);
    }
  } else {
    // equities
    transform(spec.begin(), spec.end(), spec.begin(), ::toupper);
    contract.symbol = spec;
    contract.localSymbol = spec;
    contract.secType = "STK";
    if(exchange.size())
      contract.exchange = exchange;
    else
      contract.exchange = "SMART";
    if(currency.size())
      contract.currency = currency;
    else
      contract.currency = "USD";
  }

  request_id++;
  have_details = false;
  request_new();
  _pclient->reqContractDetails(request_id, contract);
  _logger->log(Logger::INFO, "resolveContract: sent request %d for %s", request_id, contract.localSymbol.c_str());

  while(requestsOutstanding()) {
    try {
      processReplies(false);
    } catch (runtime_error& e) {
      _logger->log(Logger::ERROR, "resolveContract: %s", e.what());
      return false;
    }
  }

  if(!have_details) {
    _logger->log(Logger::WARN, "resolveContract: request %d failed", request_id);
    return false;
  }

  _logger->log(Logger::INFO, "resolveContract: request %d for %s complete", request_id, contract.localSymbol.c_str());
  return true;
}

void 
IBFetch::historicalData(TickerId reqId, const IBString& date, double open, double high,
			double low, double close, int volume, int barCount, double WAP, int hasGaps) {

  if(date.find("finished") != string::npos) {
    csv.close((csvWriter::handle)reqId);
    request_done();
    _logger->log(Logger::INFO, "historicalData: request %d complete", (int)reqId);
    return;
  }

  csv.addData((csvWriter::handle)reqId, 'B');
  csv.addData((csvWriter::handle)reqId, date);
  csv.addData((csvWriter::handle)reqId, open);
  csv.addData((csvWriter::handle)reqId, high);
  csv.addData((csvWriter::handle)reqId, low);
  csv.addData((csvWriter::handle)reqId, close);
  csv.addData((csvWriter::handle)reqId, volume);
  csv.addData((csvWriter::handle)reqId, barCount);
  csv.addData((csvWriter::handle)reqId, WAP);
  csv.addData((csvWriter::handle)reqId, hasGaps);
  csv.writeData((csvWriter::handle)reqId);
  nrows++;
}

bool
IBFetch::saveHistoricalData(const Contract& c, const string& starttime, int ndays, ib::BarSize bsize, ib::DataType dtype) {
  stringstream duration;
  duration << ndays << " D";  

  // figure out start and end times
  time_t starttime_t;
  string endtime;
  try {
    starttime_t = ibtime_to_timet(starttime);
    endtime = timet_to_ibtime(starttime_t + (ndays * 24*60*60));
  } catch (runtime_error& e) { 
    _logger->log(Logger::ERROR, "saveHistoricalData: %s", e.what());
    return false;
  }

  string fname = csv.genFilename(c, starttime);
  _logger->log(Logger::INFO, "saveHistoricalData: output to %s", fname.c_str());
  csvWriter::handle h = csv.create(fname);

  // this comment line breaks R's csv reader by default, but you can specify skip=1 to get around it.
  //
  stringstream ss;
  ss << "starttime=" << starttime << ","
     << "endtime=" << endtime << "," 
     << "duration=" << duration.str() << "," 
     << "barsize=" << ib::barsize(bsize) << "," 
     << "datatype=" << ib::datatype(dtype);
  csv.addComment(h, ss.str());

  csv.addHeader(h, "type");
  csv.addHeader(h, "datetime");
  csv.addHeader(h, "open");
  csv.addHeader(h, "high");
  csv.addHeader(h, "low");
  csv.addHeader(h, "close");
  csv.addHeader(h, "volume");
  csv.addHeader(h, "barCount");
  csv.addHeader(h, "WAP");
  csv.addHeader(h, "hasGaps");
  csv.writeHeader(h);

  request_new();
  nrows = 0;
  _pclient->reqHistoricalData((TickerId)h, c, endtime, duration.str(), ib::barsize(bsize), ib::datatype(dtype), ib::ALL, ib::STRING);
  _logger->log(Logger::INFO, "saveHistoricalData: request %d for %s, start=%s, end=%s", h, c.localSymbol.c_str(), starttime.c_str(), endtime.c_str());

  while(requestsOutstanding()) {
    try {
      processReplies(false, 10);
    } catch (runtime_error& e) {
      request_fail();
      _logger->log(Logger::ERROR, "saveHistoricalData: %s", e.what());
      return false;
    }
  }

  _logger->log(Logger::INFO, "saveHistoricalData: request %d returned %d rows", h, nrows);
  status_ss << "request returned " << nrows << " rows." << endl;
  return nrows>0;
}

void 
IBFetch::tickPrice(TickerId tickerId, TickType field, double price, int canAutoExecute) {
  RealtimeMapIterator it = rtmap.find(tickerId);
  if(it == rtmap.end())
    return;

  RealtimeContainer* rtc = it->second;
  switch(field) {
  case BID:
    rtc->bid = price;
    break;
  case ASK:
    rtc->ask = price;
    break;
  case HIGH:
    rtc->high = price;
    break;
  case LOW:
    rtc->low = price;
    break;
  case OPEN:
    rtc->open = price;
    break;
  case CLOSE:
    rtc->close = price;
    break;
  case LAST:
    rtc->epx = price;
    break;
  }
}

void 
IBFetch::tickSize(TickerId tickerId, TickType field, int size) {  
  RealtimeUpdate reason = RT_NONE;

  RealtimeMapIterator it = rtmap.find(tickerId);
  if(it == rtmap.end())
    return;
  RealtimeContainer* rtc = it->second;

  switch(field) {
  case BID_SIZE:
    rtc->bidsize = size;
    reason = RT_TOUCH;
    break;
  case ASK_SIZE:
    rtc->asksize = size;
    reason = RT_TOUCH;
    break;
  case LAST_SIZE:
    rtc->esize = size;
    reason = RT_EXECUTION;
    break;
  case VOLUME:
    rtc->volume = size;
    break;
  }
  
  writeRealtimeLine(reason, rtc);
}

void
IBFetch::writeRealtimeLine(RealtimeUpdate reason, RealtimeContainer* rtc) {
  if(reason==RT_NONE)
    return;
  if(reason==RT_TOUCH && (rtc->bid==0 || rtc->ask==0))
    return;

  string now;
  gettimestamp(now);

  if(reason == RT_TOUCH) {
    csv.addData(rtc->csvhandle, "Q");
    csv.addData(rtc->csvhandle, now);
    csv.addData(rtc->csvhandle, rtc->bid);
    csv.addData(rtc->csvhandle, rtc->ask);
    csv.addData(rtc->csvhandle, rtc->bidsize);
    csv.addData(rtc->csvhandle, rtc->asksize);
  } else if(reason == RT_EXECUTION) {
    csv.addData(rtc->csvhandle, "E");
    csv.addData(rtc->csvhandle, now);
    csv.addData(rtc->csvhandle, rtc->epx);
    csv.addData(rtc->csvhandle, rtc->esize);
  }
  csv.writeData(rtc->csvhandle);
  if(!(++realtimeCounter % 1000))
    _logger->log(Logger::INFO, "realtime: %lu events", realtimeCounter);
}

int
IBFetch::saveRealtimeData(const Contract& c, const string& starttime) {
  string fname = csv.genFilename(c, starttime);

  RealtimeContainer* rtc = new(RealtimeContainer);
  rtc->id = realtimeTicker;
  rtc->csvhandle = csv.create(fname);
  rtmap.insert(make_pair(realtimeTicker, rtc));

  _pclient->reqMktData(realtimeTicker, c, "", 0);
  _logger->log(Logger::INFO, "saveRealtimeData: request %d for %s", (int)realtimeTicker, c.localSymbol.c_str());

  realtimeTicker++;
  return 0;
}

void
IBFetch::showStatus() {
  cerr << status_ss.str();
  status_ss.str("");
}

void
IBFetch::connect(const string& connspec) throw (runtime_error) {
  string host;
  int port;
  parse_connspec(connspec, host, port);

  if(_pclient->eConnect(host.c_str(), port, _clientid) != true)
    throw runtime_error("Failed to connect");
}

void
IBFetch::disconnect() {
  _pclient->eDisconnect();
}

void
usage(char *argv0, int code) {
  cerr << "Usage: " << basename(argv0) << " [options] contract1 contract2 ..." << endl;
  cerr << " Contract may be something like USD.JPY, MSFT, ESH0" << endl << endl;
  cerr << " Options:" << endl;
  cerr << "  -l ndays: How many days worth of data to request" << endl;
  cerr << "  -s date: Data start date; follows YYYYMMDD or YYYYMMDD HH:DD:MM format" << endl;
  cerr << "  -S Step: ";
  for(int s=1; s<(int)ib::S_LAST; s++)
    cerr << ib::barsize((ib::BarSize)s) << " ";
  cerr << endl;
  cerr << "  -T Type: ";
  for(int t=1; t<(int)ib::DT_INVALID; t++)
    cerr << ib::datatype((ib::DataType)t) << " ";
  cerr << endl;
  cerr << "  -t timeout: Maximum timeout in seconds to wait for responses" << endl;
  cerr << "  -c host:port: Where to connect to" << endl;
  cerr << "  -r: Record data in realtime (tick by tick) mode" << endl;
  cerr << "  -i: client id to use; default is to use a random id" << endl;
  cerr << "  -e: exchange" << endl;
  cerr << "  -C: currency" << endl;
  cerr << "  -L logfile; the log file to use" << endl;
  exit(code);
}

int
main(int argc, char **argv) {
  ib::init();

  string connspec = "127.0.0.1:7496";
  int timeout = 1;
  int ndays = 1;
  string starttime = timet_to_ibtime(::time(NULL) - 24*60*60);
  bool realtime = false;
  string exchange;
  string currency;
  ib::BarSize barsize = ib::S_15MIN;
  ib::DataType datatype = ib::MIDPOINT;
  int clientid = 0;
  string logfile = "ibfetch.log";

  // parse options
  char c;
  while((c = getopt(argc, argv, "t:c:l:s:S:T:ri:e:C:")) != EOF) {
    switch(c) {
    case 't':
      timeout = atoi(optarg);
      break;
    case 'c':
      connspec = optarg;
      break;
    case 'l':
      ndays = atoi(optarg);
      break;
    case 's':
      starttime = string(optarg);
      break;
    case 'S':
      barsize = ib::string_to_barsize(optarg);
      break;
    case 'T':
      datatype = ib::string_to_datatype(optarg);
      break;
    case 'r':
      realtime = true;
      break;
    case 'i':
      clientid = atoi(optarg);
      break;
    case 'e':
      exchange = optarg;
      break;
    case 'C':
      currency = optarg;
      break;
    case 'L':
      logfile = optarg;
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  if(optind >= argc)
    usage(argv[0], 1);

  csvWriter csv;
  LoggerRP logger(new Logger);
  logger->init("ibfetch", logfile);

  IBFetch ib(logger);
  ib.set_clientid(clientid);
  ib.connect(connspec);
  try {
    ib.processReplies(true);
  } catch (runtime_error& e) {
    logger->log(Logger::ERROR, "%s", e.what());
    return(2);
  }

  for(int i=optind; i<argc; i++) {
    string name = argv[i];
    Contract c;
    if(ib.resolveContract(name, exchange, currency, starttime, c) == false) {
      logger->log(Logger::WARN, "could not resolve contract %s", name.c_str());
      continue;
    }
    
    if(realtime)
      ib.saveRealtimeData(c, starttime);
    else
      ib.saveHistoricalData(c, starttime, ndays, barsize, datatype);
  }

  if(realtime) {
    try {
      while(true)
	ib.processReplies(true);
    } catch (runtime_error &e) {
      logger->log(Logger::ERROR, "%s", e.what());
      return(3);
    }
  }

  ib.disconnect();
  logger->log(Logger::INFO, "Exiting, %d issues seen", ib.numIssues());
  return(ib.numIssues());
}
